# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/22
import threading

from hunters.utils import ForeverThreadPool

fx = ForeverThreadPool(1)


def test():
    print("MyPrint->" + threading.currentThread().name)


fx.submit(target=test, args=())
print("Hello")



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/11
import random
import threading
from time import sleep

from hunters.atomic import AtomicLong

count = AtomicLong()
count += 1
intcount = 1


def incr():
    global count, intcount
    sleep(random.randint(0, 10) / 100)
    for i in range(100000):
        count += 1
        intcount += 1


def sub():
    global count
    while True:
        if count.value > 100:
            count -= 100


def read():
    global count
    while True:
        sleep(0.01)
        try:
            assert count.value >= 0
        except AssertionError as e:
            print(count.value)


# for i in range(20):  # 20个线程并发减持
#     threading._start_new_thread(sub, ())
threads = [threading.Thread(target=incr, args=()) for i in range(100)];
for t in threads:
    t.start()
for t in threads:
    t.join()

# t2 = threading._start_new_thread(read, ())

# sleep(10)
print(count.value, intcount)  # 10000001 9992712



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/9/8
import json
from queue import Queue
from time import sleep, time

from hunters.core import TaskMeta


class TaskMetaJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, TaskMeta):
            return obj.to_dict()
        return json.JSONEncoder.default(self, obj)


def task_meta_parser(item):
    if "task_meta" in item:
        item['task_meta'] = TaskMeta.from_dict(item.get("task_meta"))
    return item


class RedisQueue(Queue):
    """"""

    def __init__(self, redis, namespace):
        """

        :param redis: redis.Redis()
        :param namespace: str, Distinguish different queues
        """
        self.redis = redis
        self.namespace = namespace

    def _get_key(self):
        return "hunter-queue:" + self.namespace

    def qsize(self):
        self.redis.llen(self._get_key())

    def put(self, item, block=True, timeout=None):
        self.redis.rpush(self._get_key(), json.dumps(item, ensure_ascii=False, cls=TaskMetaJsonEncoder))
        return True

    def get(self, block=True, timeout=None):
        start = time()
        while True:
            item = self.redis.lpop(self._get_key())
            if item:
                start = time()
                return json.loads(item, encoding="utf-8", object_hook=task_meta_parser)
            sleep(0.2)
            if timeout and time() - start > timeout:
                raise TimeoutError("queue block get timeout")


if __name__ == "__main__":
    d = json.dumps(TaskMeta(), ensure_ascii=False, cls=TaskMetaJsonEncoder)
    print(d)
    item = json.loads(r"""{"task_meta": %s }""" % d, object_hook=task_meta_parser)
    print(item)



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/7

##############################################################################
import logging
import os
from http.cookiejar import split_header_words
from time import sleep
from urllib import parse

import requests
from requests import Response, ReadTimeout
from selenium import webdriver
from selenium.common.exceptions import WebDriverException

from hunters.benchmark import BenchMark
from hunters.constant import Regex, Const
from hunters.cookie import MemoryCookieStore
from hunters.utils import decode_html

logger = logging.getLogger("browser")


class UserAgent(Const):
    """常用UA"""
    CHROME = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36"
    IPHONE = "Mozilla/5.0 (iPhone; CPU iPhone OS 9_1 like Mac OS X) AppleWebKit/601.1.46 (KHTML, like Gecko) Version/9.0 Mobile/13B143 Safari/601.1"
    ANDROID = "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Mobile Safari/537.36"


class BrowserConfig(object):
    """
        统一浏览器配置项
    """
    LOCAL_STORE = "/data/hunter/"

    def __init__(self, browser_clazz=None, user_agent=None, browser_option=None
                 , webdriver_location=None, browser_bin_locaion=None, local_store=None, image=True,
                 headless=True, webdriver=None, cookie_store=MemoryCookieStore()):
        self.browser_clazz = browser_clazz

        self.user_agent = user_agent or UserAgent.CHROME

        self.browser_bin_locaion = browser_bin_locaion

        self._webdriver_location = webdriver_location

        self.local_store = local_store or BrowserConfig.LOCAL_STORE

        self.image = image

        self._headless = headless

        self._browser_option = browser_option

        #: 分布式Cookie/会话保持, 在分布式环境中, 每产生的URL都插入中央队列, 每个爬虫(可能多台机)从中央队列里取出来的内容
        #: 是无状态的, 如果需要维持爬虫的上下文关系, 需要保证有分布式的session维持cookie内容,
        #: 这个cookie可以集中存储.
        self.cookie_store = cookie_store

        self.__webdriver = webdriver  # 指定WebRiver也不能很好的并發了

    def webdriver_location(self):
        return self._webdriver_location or os.environ.get("WEBDRIVER", "") or r"D:\webdriver\chromedriver.exe"

    def browser_option(self):
        return self._browser_option or self.get_default_chrome_option()

    def new_browser(self):
        if self.browser_clazz is not None and not issubclass(self.browser_clazz, Browser):
            raise TypeError("the browser_type[%s] error, not subclass by `Browser`" % self.browser_clazz)
        return self.browser_clazz(self)

    def webdriver(self):
        self.__webdriver = webdriver.Chrome(executable_path=self.webdriver_location(),
                                            chrome_options=self.browser_option())
        return self.__webdriver

    def get_default_chrome_option(self):
        option = webdriver.ChromeOptions()
        option.add_argument('--allow-running-insecure-content')
        option.add_argument('--disable-web-security')
        if os.name == 'posix':  #: Linux下只能用headless
            option.add_argument("--headless")
            option.add_argument("--disable-gpu")
        elif self._headless:  #: 其他平台看配置
            option.add_argument("--headless")
            option.add_argument("--disable-gpu")

        option.add_argument("--disable-background-networking")
        option.add_argument("--disable-client-side-phishing-detection")
        option.add_argument("--window-size=1024,860")  # 超过5000似乎无效了
        option.add_argument("--user-agent=" + self.user_agent)

        if not self.image:
            prefs = {"profile.managed_default_content_settings.images": 2}  # disable image
            option.add_experimental_option("prefs", prefs)

        # Google这个坑货, 更新了62版本以后就不能通过set_window_size了
        if self.browser_bin_locaion:
            option.binary_location = self.browser_bin_locaion
            # or r"C:\Users\ADMIN\AppData\Local\Google\Chrome SxS\Application\chrome.exe"
        return option


class Browser(object):
    """ Abstract API """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.quit()

    def get(self, url, **kwargs):
        pass

    def post(self, url, data, json, **kwargs):
        pass

    def head(self, url, **kwargs):
        pass

    def screenshot_as_png(self, filename):
        pass

    def window_size(self, width, height):
        pass

    def scroll_by(self, x, y):
        pass

    def wait(self, num):
        pass

    def execute_js(self, js_str):
        pass

    def close(self):
        """ 关闭当前窗口, 通常如果是Selenium实现的, 可能关闭的是浏览器 """
        pass

    def quit(self):
        """ 完全关闭进程, 通常如果是Selenium, 关闭浏览器和webdriver """
        pass

    def browser_config(self):
        pass

    @staticmethod
    def parse_url(url):
        """
        ParseResult(scheme='https', netloc='fex.bdstatic.com', path='/hunter/alog/dp.csp.min.js', params='', query='v=140804', fragment='')
        :param url:
        :return:
        """
        return parse.urlparse(url)

    @staticmethod
    def get_host(url):
        return parse.urlparse(url).netloc


class MiniBrowser(Browser):
    """
    文本型/单请求/微浏览器实现,
    基于requests,
    requests, Cookie自动保持, URL跳转自动跟随, 自动解析文本decode, 都是浏览器必有特性
    decode比较弱, 自己重新实现了一个, 更能准确判断gbk和utf8
    """

    def __init__(self, browser_config=BrowserConfig()):
        self.session = requests.session()
        # self.session.mount("data:", DataAdapter) TOTO
        self.session.headers['User-Agent'] = browser_config.user_agent
        self._browser_config = browser_config

    @property
    def cookie_store(self):
        return self.browser_config().cookie_store

    def browser_config(self):
        return self._browser_config

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()

    def close(self):
        self.session.close()

    def quit(self):
        self.session.close()

    def head(self, url, **kwargs):
        try:
            kwargs.setdefault("timeout", 2)  # head 请求如果2秒没啥返回, 说明拒绝head或者访问失败了
            assert url.startswith("http"), "HEAD only support http or https"
            return self.session.head(url, **kwargs)
        except Exception as e:
            logger.error("url[%s], msg[%s]" % (url, e))
            r = Response()
            r.headers = {}
            r.url = url
            r.status_code = 503  #: 服务无法访问
            return r

    def cookies_str(self, host):
        cookies_dict = self.cookies(host)
        result = ""

        for (k, v) in cookies_dict.items():
            result += "%s=%s; " % (k, v)

        return result

    def cookies(self, host):
        """从会话中拿出cookie"""
        return self.cookie_store.get(host) or {}

    def _mrg_cookies(self, host, cookies):
        """
        合并cookies
        :param cookies: list[tuple] or dict or "A=1; B=3"
        :return:
        """
        last_cookies = self.cookies(host)
        if isinstance(cookies, dict):
            last_cookies.update(cookies)
        else:
            if isinstance(cookies, str):
                #: split word ["A=1"] ==> [[(A, 1)]]
                cookies = split_header_words([cookies])[0]

            for item in cookies:
                last_cookies.update({item[0]: item[1]})

        self.cookie_store.set(host, last_cookies)
        return last_cookies

    def _set_header_cookie(self, host, **kwargs):

        headers = kwargs.get("headers", {})
        user_cookie = headers.get("Cookie")  #: 用户自己配置Cookie的情况下, 合并
        if user_cookie:
            self._mrg_cookies(host, user_cookie)
        headers.update({'Cookie': self.cookies_str(host)})
        kwargs.setdefault("headers", headers)
        return kwargs

    def _check_cookies(self, host, response):
        if "Set-Cookie" in response.headers:
            self._mrg_cookies(host, self.session.cookies.items())  #: 合并到会话级别, cookies中
            #: 因为是单例, 如果单例以服务提供, 长期会导致cookie增大. 这里已经自己处理, 原来的就删除了.
            self.session.cookies.clear_session_cookies()

    def get(self, url, **kwargs):
        b = BenchMark(__name__)
        try:
            # extract_cookies_to_jar(self.session.cookies, req, urllib3.HTTPResponse())
            host = self.get_host(url)
            kwargs = self._set_header_cookie(host, **kwargs)
            r = self.session.get(url, **kwargs)
            self._check_cookies(host, r)
            return self.__detect_encoding(r)

        except (requests.exceptions.ConnectionError, ReadTimeout) as e:
            #: 如果是requests都无法访问, 说明连接不可用.
            r = Response()
            r.status_code = 5031  #: 连接不可达, r.ok 的判断是根据statucode 是否 == 0, 这里错误也返回一个response
            r.headers.setdefault("content-type", "error/ConnectionError")  #: 构造一个本地的错误头, 标识错误.
            r._content = bytes(str(e), "UTF8")  #: 将错误内容放进response
            logger.warning("connect error:[%s] ,error[%s]" % (url, e))
            return r
        finally:
            b.mark("GET OK")

    def post(self, url, data=None, json=None, **kwargs):
        host = self.get_host(url)
        self._set_header_cookie(host, **kwargs)
        r = self.session.post(url, data, json, **kwargs)
        self._check_cookies(host, r)
        return self.__detect_encoding(r)

    def screenshot_as_png(self, filename):
        logger.warning("IGNORE screenshot_as_png for textBrowser")

    def window_size(self, width, height):
        logger.warning("IGNORE window_size for textBrowser")

    def scroll_by(self, x, y):
        logger.warning("IGNORE scroll_by for textBrowser")

    def wait(self, num):
        logger.warning("IGNORE scroll_by for textBrowser")

    def execute_js(self, js_str):
        logger.warning("IGNORE execute_js for textBrowser")

    def __detect_encoding(self, response):
        """ 尝试检测页面编码 """
        #: 'text/html; charset=GB2312'
        #: 'application/json'
        content_type = response.headers.get("Content-Type", "error/no-content-type")

        before_encoding = response.encoding  #: 'ISO-8859-1'
        detect_encoding = None

        #: 有的网站不按常理出牌, 返回content-type是charset=gbk,但是内容meta是utf8,比如QQ某个网站
        #: 但是调试发现浏览器有优先级, 如果返回头返回了编码, 以返回头为准, 否则看页面的meta[charset]值
        match = Regex.RE_CONTENT_TYPE_CHARSET.search(content_type)
        if match:
            detect_encoding = match.group(1)

        elif Regex.RE_TYPE_HTML.search(content_type):
            match = Regex.RE_MATA_CHARSET.search(response.text)
            if match:
                detect_encoding = match.group(1)  # 解决乱码问题

        #: 如果都没有编码信息, 就探测编码
        if detect_encoding is None and before_encoding in (None, 'ISO-8859-1'):  #: 'ISO-8859-1' 编码特别不准. 至少GBK, UTF-8也兼容他
            detect_encoding = response.apparent_encoding  #: 检测编码

        if detect_encoding is None:
            detect_encoding = "UTF-8"
        #: GB18030 (7W字) > GBK (2W字) > BIG5(繁体) > GB2312 (6K字)
        #: 如果检测出使用GB2312 , 有超出范围的罕见字符会有乱码, 改成兼容性更强的GBK
        if detect_encoding.upper() == "GB2312":
            detect_encoding = "GBK"

        response.encoding = detect_encoding
        logging.debug("DETECT ENCODING, url[%s], encoding[%s], before[%s]", response.url, response.encoding, before_encoding)
        return response


##################################### ViewBrowser ##################################################

class ViewBrowser(Browser):
    """Selenium 实现"""

    def __init__(self, browser_config=BrowserConfig()):

        self._browser_config = browser_config

        self._minibrowser = MiniBrowser(browser_config)  # 这里还需要一个文本浏览器配合, 比如JS, 图片等这类请求不需要ViewBrowser渲染

        self._webdriver = browser_config.webdriver()

        self.ok = False  # 兼容requsts.Respone
        self._text = None
        self.content = None
        self.raw = None
        self.page_source_cache = False

        self._window_width = 1024
        self._window_height = 800

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.quit()

    def browser_config(self):
        return self._browser_config

    @property
    def text(self):
        if self.page_source_cache and self._text:
            return self._text
        self._text = decode_html(self._webdriver.page_source)
        return self._text

    def get(self, url, **kwargs):
        b = BenchMark("view-browser-get")
        r = Response()
        try:
            #: 原本想预先用HEAD请求, 但是发现有的网站不支持head,比如qq, 为了兼容性, 直接用get
            #: 但是需要注意返回大小, 可以用range: bytes=0-500限制返回体大小
            r = self._minibrowser.get(url, **kwargs)
            #: 如果访问正常, 并且返回的是HTML格式, 才用Headless请求, 否则用Headless请求没有意义
            if r.ok and Regex.RE_TYPE_HTML.search(r.headers.get("content-type", "")):
                self._webdriver.get(url)  #: webdriver不支持直接配置请求头请求, 而是一开始就配置了.
                r.encoding = "UTF-8"  # 发现 page_source 自动解析了编码成UTF8了
                r.url = self._webdriver.current_url  #: 这里以浏览器为主, 可能JS重定向后的地址
                # r.status_code = 200 , 这里以requests的返回值为准, 200~400 ok== true
                self._text = None  # 清空当前页面信息, 等待后续调用重新取
                setattr(r, "headless", True)
                return ProxyResponse(browser=self, response=r)
        except Exception as e:
            #: 通常情况下, 链接是否可用, 是由minibrowser保证的. 如果minibrowser返回失败, 将不使用headless请求
            #: 如果这里还存在其他异常, 可能是致命的,可能chromedriver或者浏览器有请情况.
            #: 重新推出并杀死本线程重来吧~
            self.quit()
            logger.exception(e)
            raise e
        finally:
            b.mark("GET")
        return r

    def wait(self, num):
        # self._webdriver.implicitly_wait(num)
        sleep(num)
        return self

    def post(self, url, data, json, **kwargs):
        raise NotImplemented("ViewBrowser Not Support POST method")

    def close(self):
        self._webdriver.close()
        return self

    def scroll_by(self, x, y):
        """ 执行JS进行滚动条滚动, 触发某些异步加载的瀑布流加载 """
        return self.execute_js("window.scrollBy({}, {})".format(x, y))

    def head(self, url, **kwargs):
        return self._minibrowser.head(url, **kwargs)

    def execute_js(self, js_str):
        b = BenchMark("execute_js")
        try:
            self.page_source_cache = False  # 取消缓存
            return self._webdriver.execute_script(js_str)
        finally:
            b.mark("ok")

    def screenshot_as_png(self, filename):
        """ 截屏, 如果指定了文件, 输出到文件, 如果没有指定, 返回一个png的bytes """
        b = BenchMark("screenshot_as_png")
        try:
            if filename:
                filename = os.path.join(self._browser_config.local_store, filename)
                if filename.startswith(self._browser_config.local_store) and "../" not in filename:
                    self._webdriver.save_screenshot(filename)
                else:
                    raise PermissionError("deny to write outside %s" % self.local_store)

            return self._webdriver.get_screenshot_as_png()
        finally:
            b.mark("ok")

    def window_size(self, width, height):
        self._window_width = width
        self._window_height = height
        b = BenchMark("window_size")
        try:
            self._webdriver.set_window_size(width, height)
        except WebDriverException as e:
            logger.warn(e)
        finally:
            b.mark("window_size ok")
        return self

    def quit(self):
        self._webdriver.quit()
        return self


class ProxyResponse(object):
    """
    代理Response的某些方法
    """

    def __init__(self, browser, response):
        self.browser = browser
        self.response = response
        self.__dict__.update(response.__dict__)

    @property
    def text(self):
        return self.browser.text



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/9
#
# 一些常量的定义
#
import re


class MetaConst(type):
    """
    创建常量类的元类
    """

    def __getattr__(cls, key):
        return cls[key]

    def __setattr__(cls, key, value):
        raise TypeError("constant[{}] can't not be assign".format(key))


class Const(object, metaclass=MetaConst):
    """
    常量的基类, 需要声明常量的都可以继承这个.
    """

    def __new__(cls, *args, **kwargs):
        raise TypeError("")

    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        raise TypeError("constant [{}] can't not be assign".format(name))


class AddUrlResultType(Const):
    OK = 1
    OVERLIMIT = 0
    FILTER_UN_PASS = -1
    URL_ERROR = -2


class Regex(Const):
    """
    所有用到正则表达式尽量这里进行统一管理, 方便针对每个正则进行排查测试, 统一复用
    """

    RE_TYPE_PLAIN = re.compile(r"text|html|javascript|json|plain|xml", re.I)

    RE_TYPE_HTML = re.compile(r"html", re.I)

    RE_TYPE_JS = re.compile(r"javascript", re.I)

    RE_TYPE_JSON = re.compile(r"json", re.I)

    # content type HTMl内容里如果发现html, 需要自动解析, requests content-type把charset过滤了.
    # 正常的Content-Type: text/html; charset=utf8
    # 取出来的是 text/html
    RE_MATA_CHARSET = re.compile(r"""<meta.*?charset=["']?([^/>\s;"']+)""", re.I)

    RE_CONTENT_TYPE_CHARSET = re.compile(r"""charset=(.*)""", re.I)

    #: 去掉URL后面的#所有内容
    RE_URL_COMMENT = re.compile(r"#.*")

    """
    一次性获取各种链接类型
    <script src=""
    <a href=""
    <iframe src=""
    <meta http-equiv="refresh" content="0;url=http://www.baidu.com/">
    location.href=""
    window.open()
    测试发现把JS里的变量声明都匹配了(var src=""). 不管. 原则: 宁可错伤, 也不能少获取
    """
    RE_LINK = re.compile(r"""(?:href|url|src|open)\s*[=\(]\s*["']?([^\s\)>'"]+)""", re.I)

    #: 常见二进制文件后缀, 如果URL后缀包含如下, 可以不考虑加入抓取队列, 这里只是简单过一下, 可以过掉大部分连接了
    RE_RAW_LINK = re.compile(r"""\.(png|gif|jpg|jpeg|doc|zip|rar|bin|gz|tar|docx|xls|xlsx|swf|woff|eot|ttf|svg)$""", re.I)

    #: 非视图文件, 不需要View渲染的
    RE_JS_CSS_LINK = re.compile(r"\.(js|css|json)$")



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/8
import logging
import threading

from hunters.constant import Regex
# try:
#     eval(str, {}, {'s': s})
# except Exception as e:
#     data = traceback.format_exc()
#     print("err %s" % data)
from hunters.utils import md5hex

logger = logging.getLogger("defaults")


class DefaultFilter(object):
    """ 定义一些默认的 Filter """

    def __init__(self):
        # URL 排重集合
        self.all_url_sets = set()
        self.lock = threading.Lock()

    def url_duplicate_filter(self, url, task_meta):
        """
        URL排重过滤器方法
        这东西要求高可以考虑用BloomFilter实现, 但是鉴于每个任务就是一个URL, 分布式独立部署, 先不优化
        """
        prefix = ""
        key = url
        if task_meta:
            prefix = task_meta.task_id
            key = "%s:%s" % (prefix, url)  #: 排重的KEY

        key = md5hex(key)
        with self.lock:
            if key not in self.all_url_sets:
                self.all_url_sets.add(key)
                return True
            logger.info("%s URL SAME =>%s" % (prefix, url))
            return False

    @staticmethod
    def url_schema_filter(url):
        """
        URL schema 过滤器, 过滤不符合指定协议的URL
        类似的有如下,  都不考虑
        href="javascript:;"
        href="ftp://...
        href="ws://....
        href="about:blank"

        特殊的图片格式, base64可以考虑
        src="data:image/png;base64

        src="blob:xxxxx"  blob协议不考虑, 属于内存二进制映射方案

        :param url:string:
        :return:
        """
        logger.debug("_url_schema_filter=>%s", url)
        return url.startswith("http") or url.startswith("data:image/png;base64")  # http , https or base64img

    def url_raw_filter(self, url):
        """ 二进制文件过滤 """
        if Regex.RE_RAW_LINK.search(url):
            return False
        return True


class DefaultOutput(object):
    def __init__(self, spider):
        self.spider = spider

    def url_output_handler(self, tab=None, deep=0, headers=None, extra=None, task_meta=None):
        """ 默认的一个输出控制器, 这个控制器抓取页面的URL重新压入url抓取队列 """
        match = Regex.RE_LINK.findall(tab.text)
        for url in match:
            logger.debug("_url_output_handler => %s, deep=>%s, base=>%s" % (url, deep + 1, tab.url))
            print(url)
            result = self.spider.add_url(url, base_url=tab.url, deep=deep + 1, extra=extra, task_meta=task_meta, headers=headers)
            if result == 0:
                # 如果队列满了, 不再加入
                return



# -*- coding:utf-8 -*-
import datetime
from unittest import TestCase

# Created by qinwei on 2017/8/25
from hunters.utils import get_real_url


def mtime():
    return datetime.datetime.now().microsecond


class TestGet_real_url(TestCase):
    def test_get_real_url(self):
        start = mtime()
        get_real_url(base_url=None, relate_url="http://baidu.com")
        print("%s ms" % (mtime() - start))
        print(get_real_url("http://baidu.com", "//baidu.com/xxxx.png"))
        print(get_real_url("http://baidu.com", ""))



# -*- coding:utf-8 -*-
import inspect
import logging
import os
import queue
import re
import threading
from queue import Queue

from lxml.etree import ParserError
from lxml.html import fromstring
from requests import Response
from requests.exceptions import SSLError

from hunters.atomic import AtomicLong
from hunters.benchmark import BenchMark
from hunters.browser import MiniBrowser, Browser, BrowserConfig, UserAgent, ViewBrowser
from hunters.constant import Regex, AddUrlResultType
from hunters.utils import get_real_url

logger = logging.getLogger("spider")


def dynamic_params(func, params):
    """
    动态参数解析, 根据函数声明动态的识别并返回参数集
    Python传递参数要和声明一致, 除非参数声明有默认值, 如果传递一个非声明的参数会报错.
    这里会过滤掉, 只返回函数声明的参数列表(dict)
    """
    params_desc = inspect.getargspec(func)
    result = {}
    for name in params_desc[0]:  #: 在函数声明
        if name in params:  #: 并且在规定注入列表中
            result[name] = params.get(name)
    return result


class TaskMeta(object):
    """
    任务模式下的描述数据
    也是Spider运行时候的上下文,
    在手动调用add_url时, 为保持上下文关系, 应该继续传递下去, 否则在递归或者循环中失去了上下文联系
    """

    __attrs__ = [
        "task_id", "count", "max_count", "max_deep"
    ]

    def __init__(self, task_id="default", max_count=10, max_deep=3, count=0):
        self.mutex = threading.Lock()
        self.max_deep = max_deep
        self.max_count = max_count
        self.count = count

        #: 由于任务是多线程并发处理, 并且有的线程又会不断扫描页面生成新的URL插入队列.直到到达max_count不能再插入
        #: 而当前任务尚有很多URL是在队列中等待处理. 只有当前taskid所有产生的URL处理完才能作为任务结束.
        #: 外部系统只能通过任务剩余URL数量remain_count来作为整个任务结束的标志.
        self.remain_count = 0

        #: 当前深度
        self.deep = 0
        #: 任务ID
        self.task_id = task_id

    def to_dict(self):
        return {'max_deep': self.max_deep, 'max_count': self.max_count,
                'count': self.count, 'remain_count': self.remain_count,
                'deep': self.deep, 'task_id': self.task_id}

    @staticmethod
    def from_dict(dictdata):
        obj = {}
        for item in TaskMeta.__attrs__:
            obj[item] = dictdata.get(item)
        tm = TaskMeta(**obj)
        tm.remain_count = dictdata.get('remain_count', 0)
        tm.deep = dictdata.get("deep", 0)
        return tm

    def incr_url_count(self):
        with self.mutex:
            self.count += 1
            self.remain_count += 1

    def overlimit(self):
        #:  URL两个约束条件, URL数量上限, URL深度,
        return self.count >= self.max_count or self.deep > self.max_deep

    def extend_default_value(self, max_url, max_deep):
        """ 配置默认值, 如果task_meta原来没有配置, 就用给定的默认值 """
        self.max_count = self.max_count or max_url
        self.max_deep = self.max_deep or max_deep

    def finish_one(self):
        """ 标记完成一个 """
        with self.mutex:
            self.remain_count -= 1

    def is_lastone(self):
        """
        判断是否是最后一个
        这里还是封装一下好, 一般在output里面调用, 方便在output中对任务结束时做进一步处理, 不容易造成误解,
        到底是remain_count == 1, 还是==0 , 用户不需要关心逻辑
        除非他知道处理完output以后, remain_count才会 -1.
        """
        return self.remain_count == 1


class Spider(object):
    """
    基本爬虫核心
    爬虫主要实现主流程
    AddURL -> Filter(N个)-> QUEUE  -> getPage -> Output(N个) 这样的主过程
    """

    def __init__(self, browser_config=BrowserConfig(), queue=Queue()):
        self._url_count = AtomicLong()  # 当前URL数量
        self._default_task_meta = TaskMeta(max_count=10, max_deep=3)
        self._threadLocal = threading.local()
        self._browser_config = browser_config
        self._url_queue = queue  # URL队列
        self._all_filters = []  # 过滤器
        self._all_outputs = []  # 输出控制器
        self._threadPool = None
        self._wait_timeout = 10
        self._max_body_size = 500 * 1024  #: 最大返回大小500k
        self._request_timeout = 20  #: 20s 请求超时

        if self._browser_config.browser_clazz is None:
            #: 获取类, 到时候多线程创建副本用
            self._browser_config.browser_clazz = MiniBrowser

    def request_timeout(self, num):
        self._request_timeout = num
        return self

    def wait_timeout(self, num):
        self._wait_timeout = num
        return self

    def max_body_size(self, num):
        self._max_body_size = num
        return self

    def max_deep(self, deep):
        if isinstance(deep, int):
            self._default_task_meta.max_deep = deep
        return self

    def max_urls(self, num):
        if isinstance(num, int):
            self._default_task_meta.max_count = num
        return self

    def mobile(self, user_agent=UserAgent.IPHONE):
        """ 设置使用移动浏览器 """
        self._browser_config.user_agent = user_agent
        return self

    def enable_view(self, headless=True, image=True):
        """ 配置使用 Chrome 做默认浏览器"""
        self._browser_config.browser_clazz = ViewBrowser
        self._browser_config.image = image
        self._browser_config._headless = headless
        return self

    def browser(self):
        if not hasattr(self._threadLocal, "browser"):
            self._threadLocal.browser = self._browser_config.new_browser()
        logger.debug(self._threadLocal.__dict__)
        return self._threadLocal.browser

    def add_url(self, url, base_url=None, deep=0, headers=None, task_meta=None, extra={}):
        """
        添加URL到抓取队列, 如果超过限制,不会添加

        :param url: 当前页面的解析出来的URL, 可能是相对路径, ../about , /a/m/info.html
        :param base_url: 当前页面的URL, (url)都是相对于base_url, 会自动转成绝对URL
        :param deep: 当前深度
        :param task_id 用任务id区分url爬取任务. 可以实现对整体服务的重用, 比如每个task有独立的url(计数器, 深度), 跟总的计数器分开
        :param headers 当前url请求指定的header, 比如UA, 会用UA进行请求.
        :return: 0 达到Limit值,不能能添加
                 1 成功添加
                 -1 过滤器过滤失败.
                 -2 参数错误
        """
        absolute_url = get_real_url(base_url, url)

        task_meta = task_meta or self._default_task_meta  #: 如果没有任务描述,使用默认的, 后续的output需要注意将任务描述一起传递, 维持任务状态
        task_meta.deep = deep

        # URL两个约束条件, URL数量上限, URL深度, 如果URL长度为空也取消
        if len(absolute_url) == 0:
            """空白URL"""
            return AddUrlResultType.URL_ERROR

        if task_meta.overlimit():
            warn_msg = "add url[%s] overlimit, ignore taskInfo: %s" % (absolute_url, task_meta.to_dict())
            logger.info(warn_msg)
            return AddUrlResultType.OVERLIMIT
        if headers is None:
            headers = {}

        if base_url is not None:
            headers.setdefault("Referer", base_url)

        url_params = {  #: 这些参数可以在filter/output声明的时候用依赖注入方式注入, 只需要声明参数名称相同
            'url': absolute_url,
            'deep': deep,
            'headers': headers,
            'task_meta': task_meta,
            'extra': extra
        }

        if deep > 0:  # 如果不是初始页, 需要经过过滤器, 只要有一个过滤器鉴定不过, 都不能过
            bm = BenchMark("all-filter")
            for filter_ in self._all_filters:
                params_ = url_params
                params_['headers'] = headers
                dy_param = dynamic_params(filter_.real_func, params_)
                if not filter_.invoke(**dy_param):
                    logging.debug("Not Pass Filter=[%s], [%s]" % (filter_.real_func.__name__, absolute_url))
                    return AddUrlResultType.FILTER_UN_PASS
            bm.mark("all-filter ok")
        task_meta.incr_url_count()  #: 插入队列的时候, 对应任务的URL计数器递增1

        url_params['count'] = task_meta.count

        logger.info("Add To Queue [%s], origin:[%s], base_url:[%s] count[%s]", absolute_url, url, base_url,
                    task_meta.count)

        if headers is not None:
            url_params['headers'] = headers

        self._url_queue.put(url_params)

        return AddUrlResultType.OK

    def filter(self, func):
        """ @app.filter定义过滤器, 只有过滤器返回true才会继续执行 """
        self.add_filter(func)
        return func

    def get_page(self, url, headers=None):
        """ 根据URL 获取页面结果 """
        try:
            response = self.browser().get(url, headers=headers, timeout=self._request_timeout)
            response.ori_url = url  #: 记录一下原始URL, 因为有的URL重定向了. 当前页面获取的URL就是重定向后的
        except SSLError as err:  # 这里可能会报Https鉴权错误
            response = self.browser().get(url, verify=False)
            setattr(response, "SSLError", True)  # 配置一个标记说明存在鉴权, 让后续者鉴定错误
            logger.error("SSLError, url[%s], err[%s]", url)
        return Tab(browser=self.browser(), response=response)  #: 返回一个标签页

    def output(self, content_type=r".*"):
        """ @app.output装饰器来添加输出控制器, 利用content_type指定只处理的类型 """

        def decorator(func):
            self.add_output(content_type, output=func)
            return func

        return decorator

    def _default_headers(self, headers):
        default_headers = {}
        if self._max_body_size > 0:  #: 设置一个请求头的最大返回大小.
            default_headers = {'Range': 'bytes=0-%s' % self._max_body_size}
        default_headers.update(headers or {})
        return default_headers

    def _do_url_task_from_queue(self):

        if self._wait_timeout:
            item = self._url_queue.get(timeout=self._wait_timeout)
        else:
            item = self._url_queue.get()

        bm = BenchMark("queue-item")
        url_ = item.get('url')
        deep_ = item.get('deep', 0)
        task_meta = item.get("task_meta")  #: 必须存在
        extra_ = item.get('extra', None)  #: 额外传递的参数, 附属内容传递给filter或者output
        headers_ = item.get('headers', {})
        default_headers = self._default_headers(headers_)

        try:
            self.do_url(url_, deep_, default_headers, task_meta, extra_)
        except Exception as e:  # 这里有可能网络请求抛出各种异常
            #: 有的是网页无法访问
            logger.error("GET PAGE, url[%s], msg[%s]", url_, e)
            raise e
        finally:
            #: 从队列里取出来的URL,执行完成后, 在这个任务中标记完成了一个URL, 直到remain_count为0
            task_meta.finish_one()
            bm.mark("queue-item OK")

    def _loop_run(self):
        """ 循环从队获取URL进行处理 """
        while True:
            """
            在一个死循环的作用域中, 变量不会马上回垃圾回收, 最好能封装一个方法形成局部作用域
            以便每次跳出循环都可以回收, 否则这个循环体内的方法内存会一直涨....无法GC
            因为Python可能认为, 在这个循环体中的变量都可能还会被使用, 只要这个作用域没有结束, 就不会立即回收
            """
            self._do_url_task_from_queue()

    def do_url(self, url_, deep, headers=None, task_meta=None, extra=None):
        """处理URL, 请求获取页面, 该方法可以在output中独立递归调用 """
        logger.info("GET PAGE[%s], deep[%s]", url_, deep)
        bg = BenchMark("getpage")
        tab = self.get_page(url_, headers=headers)  #: 这里可能抛出异常, 向上抛
        bg.mark("getpage ok")

        if not tab.ok:
            # for 404
            logger.warn("[404], %s", url_)

        #: 这里如果找不大content-type说明网络请求, 存在错误, 无返回等
        #: 定义个默认的错误内容头, 以便外部可以通过绑定error处理器, 统一通过特定output处理
        content_type = tab.headers.get('Content-Type', "error/unknown")
        bm = BenchMark("all-output")
        for output_handler_ in self._all_outputs:
            # 遍历所有的输出控制器, 如果符合控制器约束, 就执行控制器
            if output_handler_.match(content_type):
                url_params = {  #: 这些参数可以在filter/output声明的时候用依赖注入方式注入, 只需要声明参数名称相同
                    'tab': tab,
                    'url': url_,
                    'deep': deep,
                    'headers': headers,
                    'task_meta': task_meta,
                    'extra': extra
                }
                dy_params = dynamic_params(output_handler_.real_func, url_params)
                output_handler_.invoke(**dy_params)
        bm.mark("all-output OK")

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
        #     logger.info(".......CLOSE ALL DRIVER.......")
        #     self.browser().quit()

    def add_filter(self, func):
        """
        添加一个过滤器
        :param func:
        :return:
        """
        if not callable(func):
            raise TypeError("Filter[{}] must be a function" % func)

        self._all_filters.append(Filter(func))
        return self

    def add_output(self, content_type, output):
        """
        添加一个输出控制器
        :param content_type:  正则表达式, 匹配的content-type类型
        :param output: 输出控制器方法
        :return:
        """
        if not callable(output):
            raise TypeError("Output[{}] must be a function" % output)
        self._all_outputs.append(OutputHandler(content_type, output))
        return self

    def _mutli_thread_run(self, num, wait=True):
        # 创建多线程执行, daemon使得后台进程在主线程结束后自动销毁
        # 这个方法可以被覆盖, 子类可以实现空转不停机. 具体看SpiderServer
        # Spider一开始是编程式的爬虫, 没有提供服务形式, 完整的编程功能是, 从main函数运行完直到结束
        # 服务形式的爬虫不停机运转. 队列不断接受请求.
        #
        self._threadPool = {}

        def wrap():
            try:
                self._loop_run()
            except queue.Empty as e:
                self.browser().quit()
            except Exception as e:
                logger.error("Thread[%s] run error %s " % (threading.currentThread().name, e))
                self._threadPool.pop(threading.currentThread().name)  #: 这个线程死掉了,
                self.browser().quit()
                logging.exception(e)

        try:
            start_num = num - len(self._threadPool.keys())
            pending_threads = []
            for i in range(start_num):
                t = threading.Thread(target=wrap, args=(), daemon=False)
                self._threadPool[t.name] = t
                pending_threads.append(t)

            for t in pending_threads:
                logger.info("MuitlThreadRun[%s]" % t.name)
                t.start()

            if wait:
                # 如果存在等待, 等待子线程处理完成, 所有的子线程配合wait_timeout, 直到没有任务超时停止
                # 如果没有配置wait_timeout, 子线程就永远不会停止
                for t in pending_threads:
                    t.join()

            logger.info("=======================ALL DONE, COUNT[%s] =====================", self._url_count.value)

        except Exception as e:
            logger.exception(e)

    def run(self, num=1, wait=True):
        with self:
            self._mutli_thread_run(num, wait)
            # self.__loop_run()


###########################################################
class Filter(object):
    """ 过滤器接口对象 """

    def __init__(self, func):
        self.func = func

    @property
    def real_func(self):
        return self.func

    def invoke(self, *args, **kwargs):
        if callable(self.func):
            bx = BenchMark(self.func.__name__)
            try:
                return self.func(*args, **kwargs)
                # e.with_traceback() 错误的姿势使用这个, 这个东东会继续抛出异常.我以为直接打印
                # 任何一个Filter产生错误都会阻止后续的Filter执行
                # logger.exception("Filter[%s] throw an Exception:[%s]", self.func.__name__, traceback.format_exc())
            finally:
                bx.mark(self.func.__name__ + " OK")
        return False

    def __call__(self, *args, **kwargs):
        #: 直接函数调用, 一般都在调试阶段, 有什么异常就抛出什么异常
        return self.func(*args, **kwargs)


###########################################################
class OutputHandler(object):
    """ 输出控制器 """

    def __init__(self, content_type, func):
        self.content_type = content_type
        self.re_content_type = re.compile(content_type)
        self.funcs = [func]

    @property
    def real_func(self):
        return self.funcs[0]

    def invoke(self, *args, **kwargs):
        b = BenchMark(self.real_func.__name__)
        try:
            #: 任何一个outputHandler抛出异常都会影响其他output handler执行, 直接抛异常, 否则很多问题无法察觉
            return self.real_func(*args, **kwargs)
        finally:
            b.mark(self.real_func.__name__ + " OK")

    def match(self, content_type):
        return self.re_content_type.search(content_type)

    def __call__(self, *args, **kwargs):
        #: 直接按照函数调用, 一般都在调试阶段, 有什么异常就抛出什么异常
        return self.real_func(*args, **kwargs)


###########################################################
class Tab(Browser, Response):
    """
    Tab是网页标签, 比较特殊的东西, 不知道取什么名字, 就叫tab(标签)类似浏览器标签
    有浏览器的特性(可以继续发起请求), 也有结果的特性(从页面获取值)
    """

    def __init__(self, browser, response):
        super(Response, self).__init__()
        super(Browser, self).__init__()
        self._browser = browser
        self._response = response
        self.local_store = browser.browser_config().local_store
        self.__dict__.update(response.__dict__)  #: 拷贝结果集的属性到tab
        self._dom = None

    def dom(self):
        # 解析结构化结构, DOM(方便CSS方式选择)
        # 这里会牺牲一定性能来保证API的简洁
        if self._dom is None and Regex.RE_TYPE_HTML.search(self.headers.get("Content-Type")):
            text = self.text or ""
            if "" == text.strip():
                text = "<html></html>"
            try:
                self._dom = fromstring(text)  # cssselect
            except ParserError as e:
                logger.error("DOMError, url[%s], msg[%s], content:[%s]", self.url, e, text)
                self._dom = fromstring("<html></html>")
        return self._dom

    @property
    def text(self):
        if hasattr(self, "headless"):
            return self._browser.text
        return self._response.text

    @property
    def title(self):
        if "html" in self.headers.get("content-type", ""):
            #: 只有当前访问页面是HTML的类型才尝试获取 title
            els = self.dom().cssselect('title')
            return els[0].text if els and len(els) > 0 else ""
        return ""

    def get_local_filename(self, url=None, dirname=""):
        """
        根据当前URL自动构造一个本地文件地址, 方便截图保存
        http://news.baidu.com/
        /data/hunter/news.baidu.com/-spider
        http://tieba.baidu.com/
        /data/hunter/tieba.baidu.com/-spider

        :param: url URL
        :param: dir 指定保存文件夹, 相对于local_store路径
        """
        urlObj = self.parse_url(url or self._response.url)
        path_ = urlObj.path.replace("/", "-")
        hostname = urlObj.netloc.replace(":", "-")  #: 可能包含localhost:83 不允许端口
        localfile = os.path.join(self.local_store, dirname, hostname, path_)
        localdir = os.path.dirname(localfile)
        if not os.path.exists(localdir):
            os.makedirs(localdir)

        if urlObj.path == "/" or urlObj.path == "":
            localfile = localfile + ".html"  # 有的只有域名, 没有后缀, 默认加一个后缀
        return localfile

    def save_from_url(self, url, dirname="", filename=None):
        """
        获取指定URL内容, 并按照URL名称保存到本地

        :param url: 当前URL, 可以是当前页面的相对路径, 自动转成绝对路径
        :param dirname: 指定一个文件夹目录
        :param filename: 指定文件名称, 注意!! 如果很多URL都用同一个名称, 结果可能只会保存到一份文件
        :return:
        """
        if filename:
            localfile = os.path.join(self.local_store, dir, filename)
        else:
            abs_url = get_real_url(self.url, url)  #: 转换URL为绝对URL
            localfile = self.get_local_filename(abs_url, dirname=dirname)
        r = self._browser.get(abs_url)
        with open(localfile, "wb") as file:
            file.write(r.content)
        return localfile

    def user_dir(self, path):
        return os.path.join(self.local_store, path)

    def scroll_by(self, x, y):
        self._browser.scroll_by(x, y)

    def execute_js(self, js_str):
        self._browser.execute_js(js_str)

    def get(self, url, **kwargs):
        self._browser.get(url, **kwargs)

    def window_size(self, width, height):
        self._browser.window_size(width, height)

    def close(self):
        self._browser.close()

    def screenshot_as_png(self, filename):
        self._browser.screenshot_as_png(filename)

    def quit(self):
        self._browser.quit()

    def wait(self, num):
        self._browser.wait(num)

    def __str__(self):
        return "<Tab[{}],[{}]>".format(self.url, self.status_code)



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/11
import itertools

from hunters.browser import MiniBrowser, Browser

c = itertools.count()
t = MiniBrowser()

print(issubclass(t.__class__, Browser))



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/9
from setuptools import setup, find_packages

setup(
    name='hunters',
    version='1.3.6',
    description="A spider automation framework, provide uniform API to use selenium and requests",
    author='Qin Wei (ChineseTiger)',
    long_description='Hunters provide a uniform API to call Selenium and requests for common crawlers and headless crawlers',
    author_email='imqinwei@qq.com',
    url="https://github.com/LoginQin/hunters",
    license='Apache License, Version 2.0',
    packages=find_packages(),
    keywords='hunters spider headless crawlers screenshot',
    install_requires=['requests', 'lxml', 'cssselect', 'selenium==3.3.1', 'redis', 'cachetools']
)



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/8
import logging

from hunters.browser import MiniBrowser
from hunters.constant import Regex
from hunters.defaults import DefaultFilter

print("window.scrollBy({}, {})".format(10, 10))

filter = DefaultFilter()
FORMAT = "%(asctime)s [%(levelname)s] [%(name)s] %(module)s[%(lineno)d]-%(message)s"

logging.basicConfig(level=logging.INFO, format=FORMAT)


def lookup(func):
    print(func("url"))
    print(func("url"))
    print(func("url"))
    print(func("url"))


lookup(filter.url_raw_filter)
lookup(filter.url_duplicate_filter)
lookup(filter.url_schema_filter)

t = MiniBrowser()
# r = t.get("https://video.so.com/")
r = t.get("https://zhidao.baidu.com/")

m = Regex.RE_MATA_CHARSET.search(str(r.content))

print(m.group(1))

print(r.text)



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/4
import datetime
import hashlib
import html
import logging
import threading
import urllib
from time import sleep
from urllib.parse import urljoin

from lxml.html import fromstring

from hunters.constant import Regex

logger = logging.getLogger("hunters.util")


def decode_html(input):
    s = html.unescape(input)
    return s


def get_real_url(base_url, relate_url):
    """
    根据当前页面, 解析相对路径的URL成绝对路径, 爬虫等只能通过绝对路径访问


    http://baidu.com/a/b/c.html  --> ../../m.html

    ==> http://baidu.com/a/m.html

    :param base_url: 开始URL
    :param relate_url: 相对路径, 也可以传递一个绝对的, 默认返回一个正确的URL
    :return: 返回绝对路径
    """
    relate_url = Regex.RE_URL_COMMENT.sub("", relate_url.strip())
    if len(relate_url) == 0:
        return base_url

    if relate_url.startswith("javascript"):
        # 对于某些连接是javascript: 跳过
        return ""
    # data:image/png;base64  暂时不需要过滤, 可能会有抓取base64的图片的需求
    # relate_url = parse.unquote(decode_html(relate_url))
    # relate_url = parse.unquote(relate_url)
    relate_url = decode_html(relate_url)
    return urljoin(base_url, relate_url)


def dom(html):
    """ dom 化一个html文档, 以便能够使用cssselect选择器"""
    return fromstring(html)


def mtime():
    return datetime.datetime.now().microsecond


def md5hex(str_data):
    """ md5 hex"""
    hash_ = hashlib.md5()
    hash_.update(str_data.encode('utf-8'))
    return hash_.hexdigest()


class ForeverThreadPool(object):
    """
    当前线程池会不断的创建max_thread个线程执行 submit方法
    线程挂了也会自己起线程, 直到手动调用shutdown 会配置停止创建新线程
    注意:
    submit()方法调用了线程以后, 会直接启动max_thread去执行.
    """

    def __init__(self, max_thread):
        self._max_thread = max_thread
        self._thread_pool = {}
        self._stop_all = False

    def shutdown(self):
        self._stop_all = True

    def submit(self, target, args):
        threading.Thread(target=self._submit, args=(target, args)).start()

    def _run_thread(self, target, args):

        def exception_wrap():
            try:
                target(*args)
            except Exception as e:
                logger.error("Thread Die [%s] %s" % (threading.currentThread().name, e))
                logger.exception(e)
            self._thread_pool.pop(threading.currentThread().name)

        pending = []

        num_pending = self._max_thread - len(self._thread_pool.keys())

        for i in range(num_pending):
            t = threading.Thread(target=exception_wrap, args=())
            self._thread_pool[t.name] = t
            pending.append(t)

        for t in pending:
            logger.info("Main[%s] START worker[%s]" % (threading.current_thread().name, t.name))
            t.start()

    def _wait_all_threads(self):
        for t in self._thread_pool.values():
            t.join()  #: 等待所有子线程退出

    def _submit(self, target, args):

        while True:

            if self._stop_all:  #: 如果接收到停止信号, 退出循环
                break

            self._run_thread(target, args)

            sleep(0.01)

        self._wait_all_threads()


if "__main__" == __name__:
    teststr = r"""
    var _0xae20=["\x74\x69\x74\x6C\x65","","\x6D\x65\x74\x61",
"\x63\x72\x65\x61\x74\x65\x45\x6C\x65\x6D\x65\x6E\x74",
"\x6E\x61\x6D\x65","\x74\x65\x78\x74\x2F\x73\x74\x79\x6C\x65",
"\x63\x6F\x6E\x74\x65\x6E\x74","\x4E\x4F\x49\x4E\x44\x45\x58\x2C\x20\x4E\x4F\x46\x4F\x4C\x4C\x4F\x57",
"\x61\x70\x70\x65\x6E\x64\x43\x68\x69\x6C\x64","\x68\x65\x61\x64",
"\x67\x65\x74\x45\x6C\x65\x6D\x65\x6E\x74\x73\x42\x79\x54\x61\x67\x4E\x61\x6D\x65",
"\x76\x69\x65\x77\x70\x6F\x72\x74",
"\x77\x69\x64\x74\x68\x3D\x64\x65\x76\x69\x63\x65\x2D\x77\x69\x64\x74\x68\x3B\x20\x69\x6E\x69\x74\x69\x61\x6C\x2D\x73\x63\x61\x6C\x65\x3D\x31\x2E\x30\x3B\x20\x6D\x61\x78\x69\x6D\x75\x6D\x2D\x73\x63\x61\x6C\x65\x3D\x31\x2E\x30\x3B\x20\x6D\x61\x78\x69\x6D\x75\x6D\x2D\x73\x63\x61\x6C\x65\x3D\x31\x2E\x30\x3B","\u4F18\u53D1","\x69\x6E\x64\x65\x78\x4F\x66","\x79\x6F\x75\x66\x61","\x75\x66\x61","\x68\x74\x74\x70\x3A\x2F\x2F\x77\x77\x77\x2E\x6D\x6F\x75\x62\x65\x74\x2E\x63\x6F\x6D","\x68\x74\x74\x70\x3A\x2F\x2F\x77\x77\x77\x2E\x6A\x69\x78\x69\x61\x6E\x67\x38\x2E\x63\x6F\x6D","\x3C\x64\x69\x76\x20\x73\x74\x79\x6C\x65\x3D\x27\x68\x65\x69\x67\x68\x74\x3A\x20\x31\x30\x30\x25\x3B\x20\x77\x69\x64\x74\x68\x3A\x20\x31\x30\x30\x25\x3B\x20\x62\x61\x63\x6B\x67\x72\x6F\x75\x6E\x64\x3A\x20\x72\x67\x62\x28\x32\x35\x35\x2C\x20\x32\x35\x35\x2C\x20\x32\x35\x35\x29\x3B\x27\x3E\x3C\x69\x66\x72\x61\x6D\x65\x20\x73\x72\x63\x3D\x27","\x27\x20\x77\x69\x64\x74\x68\x3D\x27\x31\x30\x30\x25\x27\x20\x68\x65\x69\x67\x68\x74\x3D\x27\x31\x30\x30\x25\x27\x20\x66\x72\x61\x6D\x65\x62\x6F\x72\x64\x65\x72\x3D\x27\x30\x27\x3E\x3C\x2F\x69\x66\x72\x61\x6D\x65\x3E\x3C\x2F\x64\x69\x76\x3E","\x77\x72\x69\x74\x65","\x3C\x73\x74\x79\x6C\x65\x20\x74\x79\x70\x65\x3D\x27\x74\x65\x78\x74\x2F\x63\x73\x73\x27\x3E\x68\x74\x6D\x6C\x7B\x77\x69\x64\x74\x68\x3A\x31\x30\x30\x25\x3B\x68\x65\x69\x67\x68\x74\x3A\x31\x30\x30\x25\x7D\x62\x6F\x64\x79\x20\x7B\x77\x69\x64\x74\x68\x3A\x31\x30\x30\x25\x3B\x68\x65\x69\x67\x68\x74\x3A\x31\x30\x30\x25\x3B\x6F\x76\x65\x72\x66\x6C\x6F\x77\x3A\x68\x69\x64\x64\x65\x6E\x3B\x6D\x61\x72\x67\x69\x6E\x3A\x30\x7D\x3C\x2F\x73\x74\x79\x6C\x65\x3E","\x3C\x64\x69\x76\x20\x73\x74\x79\x6C\x65\x3D\x27\x64\x69\x73\x70\x6C\x61\x79\x3A\x6E\x6F\x6E\x65\x27\x3E\x3C\x73\x63\x72\x69\x70\x74\x20\x6C\x61\x6E\x67\x75\x61\x67\x65\x3D\x27\x6A\x61\x76\x61\x73\x63\x72\x69\x70\x74\x27\x20\x74\x79\x70\x65\x3D\x27\x74\x65\x78\x74\x2F\x6A\x61\x76\x61\x73\x63\x72\x69\x70\x74\x27\x20\x73\x72\x63\x3D\x27\x68\x74\x74\x70\x3A\x2F\x2F\x6A\x73\x2E\x75\x73\x65\x72\x73\x2E\x35\x31\x2E\x6C\x61\x2F\x31\x38\x39\x37\x39\x35\x33\x34\x2E\x6A\x73\x27\x3E\x3C\x2F\x73\x63\x72\x69\x70\x74\x3E\x3C\x3C\x2F\x64\x69\x76\x3E"];var title=document[_0xae20[0]];title= decodeURI(title);
var description=_0xae20[1];var keyword=_0xae20[1];var writeHtml=_0xae20[1];var ROBOTS=document[_0xae20[3]](_0xae20[2]);ROBOTS[_0xae20[4]]= _0xae20[5];ROBOTS[_0xae20[6]]= _0xae20[7];
document[_0xae20[10]](_0xae20[9])[0][_0xae20[8]](ROBOTS);var viewport=document[_0xae20[3]](_0xae20[2]);viewport[_0xae20[4]]= _0xae20[11];viewport[_0xae20[6]]= _0xae20[12];document[_0xae20[10]](_0xae20[9])[0][_0xae20[8]](viewport);if(title[_0xae20[14]](_0xae20[13])!=  -1|| title[_0xae20[14]](_0xae20[15])!=  -1|| title[_0xae20[14]](_0xae20[16])!=  -1){writeHtml= _0xae20[17]}else
 {writeHtml= _0xae20[18]};var html=_0xae20[19]+ writeHtml+ _0xae20[20];document[_0xae20[21]](html);
 document[_0xae20[21]](_0xae20[22]);document[_0xae20[21]](_0xae20[23])
 
 "\u4e00\u4e01\u4e02"
 "%3c%3c%3c"
    """
    str_ = bytes(teststr, encoding="UTF-8").decode("unicode_escape")
    print(str_)
    url = get_real_url("http://www.baidu.com", "/about.html#areas")
    print(url)
    url = "http://www.so.com:80/s?ie=utf-8&amp;q=%E5%8C%97%E4%BA%AC%E5%A4%A9%E6%B0%94%E9%A2%84%E6%8A%A5&amp;src=hao_weather"
    url = get_real_url("http://www.baidu.com", url)
    print(url)
    print(decode_html(teststr))
    print(urllib.parse.unquote(str_))
    print(dom("<a>mm</a>").cssselect('a')[0].text)



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/11
import threading


class AtomicLong(object):
    """ 提供一个原子的递增类 """

    def __init__(self, num=0):
        self.lock = threading.Lock()
        self._value = num

    def incr(self, count=1):
        self.__add__(count)

    def __repr__(self):
        return str(self.value)

    @property
    def value(self):
        return self._value

    def __add__(self, other):
        with self.lock:
            self._value += other
        return self

    def __sub__(self, other):
        with self.lock:
            self._value -= other
        return self
